function [Endmembers, Abundances] = FUN_algorithm( IMG_cub, StopFactor)

% Fast UNmixing Algorithm (FUN).

% Published in: Geoscience and Remote Sensing, IEEE Transactions on, vol. PP, no. 99, pp. 114, 2015
% Title: "A New Fast Algorithm for Linearly Unmixing Hyperspectral Images."
% Authors: R. Guerra, L. Santos, S. Lopez, and R. Sarmiento

% Overall explanation. The FUN_algorithm is divided in two main steps. The
% first step estimates the number of endmembers and extracts them at the same time.
% The second step calculates the abundances of the extracted endmembers.
% Both steps can be used independently.

% INPUTS:

    % IMG_cub    -> Hyperspectral cube (lines x rows x bands)
    % StopFactor -> Factor which determines when to stop extracting endmembers.
    %               Percentage of information which will be lost in the
    %               linear reconstruction of the hyperspectral image.
    
% OUTPUTS:

    % Endmembers -> Matrix with the endmembers placed in columns
    % Abundances -> (lines x rows x number of endmembers) abundances matrix
   
% Copyright (2015) DSI IUMA @ University of Las Palmas of Gran Canaria, Spain. 
    
    
    
%% ------------------------ Algorithm ------------------------

%% 0.) Estimate the dimension of the image. Set the pixels in columns

[nr, nc, nb] = size(IMG_cub);
N = nr*nc;
IMG = reshape(IMG_cub, N, nb)';


%% 0.1) Noise filtering 

    % Any noise filter can be useful. For small images the smooth filter
    % may provide better results than the HySIME filter. (They can also be 
    % sequentially used)

[IMG] = function_NoiseFilteringHySIME(IMG);

[IMG] = function_NoiseFilteringSmooth(IMG);


%% 1.) Selecting the centroid pixel

centroid = mean(IMG')';
r0 = norm(centroid);
q0 = centroid/r0;


%% 2.) FUN first iteration. Extraction of the first endmember

IMG2 = IMG;
IMG2 = IMG2 - q0*(q0'*IMG2);
for k = 1:1:N
    IMG_error(1,k) = norm(IMG2(:,k));
end
[max_Value, max_Index] = max(IMG_error);

r1 = norm(IMG(:,max_Index));
q1 = IMG(:,max_Index)/r1;

Q = [q1];
Indexes = [max_Index];

    
%% 3.) FUN algorithm for endmember extraction

IMG2 = IMG;
exit = 0;
pmax = 50;  % Maximun number of endmembers that will be extracted even if the StopFactor value is not reached
j = 2;
while j <= pmax && exit == 0
    IMG2 = IMG2 - Q(:,j-1)*(Q(:,j-1)'*IMG2);
    for k = 1:1:N
        IMG_error(1,k) = norm(IMG2(:,k));
    end
    [max_Value, max_Index] = max(IMG_error);
    if norm(IMG2(:,max_Index))/norm(IMG(:,max_Index))*100 >= StopFactor
        Q = [Q, IMG2(:,max_Index)/norm(IMG2(:,max_Index))];
        Indexes  = [Indexes, max_Index];
        j = j + 1;
    else
        exit = 1;
    end
end

P = length(Indexes);
Endmembers = [];
for i=1:1:P
    Endmembers = [Endmembers, IMG(:,Indexes(i))];
end



%% 4.) Abundances calculation

FC = 0; % FC = 1 fully constrained abundances;
        % FC = 0 no constrained abundances;
[ A ] = function_Abundances( Endmembers, IMG, FC);
        % A = Abundances. Each line correspond to one endmember
for k = 1:1:P
    Abundances(:,:,k) = reshape(A(k,:), nr, nc);
        % Abundances (nr x nl x P). Each band correspond to one endmember
end
        




%% ------------------------ Functions ------------------------


%% Function for the abundances calculation

function [ A ] = function_Abundances( E, IMG, FC )

[nb, P] = size(E);
[nb, N] = size(IMG);

E = [E, E]; % Endmembers
rr = [];    % Norms of the orthogonalized vectors
Q = [];     % Orthogonal vectors to project the pixels of the image.
for k = 2:1:P+1
    r1 = norm(E(:,k));
    q = E(:,k)/r1;
    for j = 2:1:P
        X = E(:,k+j-1);
        for i = 1:1:j-1
            rj = X' * q(:,i);
            X = X - rj * q(:,i);
        end
        rj = norm(X);
        qj = X/rj;
        q = [q, qj];
    end
    rr = [rr, rj];
    Q = [Q, qj];
    q = [];
end
% divided by the norm
for i = 1:1:P
    Q(:,i) = Q(:,i)/rr(i);
end
A = Q'*IMG;
% Constraints
if FC == 1
    % NoNegativeConstraint
    for i = 1:1:P
        for j = 1:1:N
            if A(i,j)<0
                A(i,j) = 0; 
            end
        end
    end
    % SumToOneConstraint
    for j = 1:1:N
        A(:,j) = A(:,j)/sum(A(:,j));
    end
end
return



%% Function for the noise filtering using the HySIME filter

function [ IMG ] = function_NoiseFilteringHySIME( IMG )
r = IMG;
small = 1e-6;
[L N] = size(r);
w=zeros(L,N);
RR=r*r';     
RRi=inv(RR+small*eye(L));
for i=1:L
    XX = RRi - (RRi(:,i)*RRi(i,:))/RRi(i,i);
    RRa = RR(:,i); RRa(i)=0; % this removes the effects of XX(:,i) equation (9)
    beta = XX * RRa; beta(i)=0; % this removes the effects of XX(i,:) equation (10)
    w(i,:) = r(i,:) - beta'*r; % note that beta(i)=0 => beta(i)*r(i,:)=0
end
Rw=diag(diag(w*w'/N));
IMG = IMG - w;
return



%% Function for the noise filtering using the Smooth (matlab default) filter

function [ IMG ] = function_NoiseFilteringSmooth( IMG )
[nb N] = size(IMG);
for i = 1:1:N
    IMG(:,i) =  smooth(IMG(:,i));
end
return




